import dataset
from .metadata import ChainMetadata
from .utils import fetch_cached_file

class History:
    def __init__(self, url: str = "http://chains.mggg.org.s3-website-us-east-1.amazonaws.com/"):
        if not url.endswith("/"):
            url += "/"
        self.url = url

        fetch_cached_file(self.url, "/tmp/chains.db")

        self.db = dataset.connect('sqlite:////tmp/chains.db')
        self.chains = self.db["chains"]
    
    def search(self, **kwargs):
        for result in self.chains.find(**kwargs):
            metadata = ChainMetadata(**result)
            metadata.url = self.url
            yield metadata

    def search_one(self, **kwargs):
        result = self.chains.find_one(**kwargs)
        if result:
            metadata = ChainMetadata(**result)
            metadata.url = self.url
            yield metadata


from networkx.readwrite import generate_adjlist
import gerrychain
import json
import hashlib
import os
import requests
import shutil
from dateutil.parser import parse


def compute_graph_hash(graph) -> str:
    return hashlib.sha256(generate_adjlist("".join(graph)).encode()).hexdigest()

def fetch_cached_file(url: str, loc: str):
    fetch_file = True
    if os.path.isfile(loc):
        r = requests.head(url)
        last_modified = parse(r.headers['last-modified']).timestamp()
        file_timestamp = os.path.getmtime(loc)

        if file_timestamp >= last_modified: 
            fetch_file = False

    if fetch_file:
        with requests.get(url, stream=True) as r:
            with open(loc, 'wb') as f:
                shutil.copyfileobj(r.raw, f)


from .history import ChainMetadata
from .utils import compute_graph_hash

from gerrychain import Partition, GeographicPartition
from typing import Union
import os
import json
import requests
import shutil
import multiprocessing

import subprocess

class Replay:
    def __init__(
        self,
        graph,
        chain_name: Union[str, ChainMetadata],
        updaters=None,
        executable="pcompress -d",
        # executable="pv",
        threads=None,
        geographic=False,
        flips=True,
        *args,
        **kwargs,
    ):
        self.graph = graph

        if isinstance(chain_name, ChainMetadata):
            self.filename = f"/tmp/{ChainMetadata.identifier}.chain"
            if not os.path.isfile(self.filename):
                with requests.get(chain_name.url + chain_name, stream=True) as r:
                    with open(self.filename, 'wb') as f:
                        shutil.copyfileobj(r.raw, f)
            assert compute_graph_hash(self.graph) == chain_name.graph_hash, "Not the same graph!"
        else:
            self.filename = chain_name

        self.updaters = updaters
        self.geographic = geographic
        self.flips = flips
        self.executable = executable

        self.args = args
        self.kwargs = kwargs

        if not threads:
            self.threads = multiprocessing.cpu_count()
        else:
            self.threads = threads

        if self.flips and self.executable == "pcompress -d":
            self.executable = "pcompress -d --diff"

        self.child = subprocess.Popen(
            f"/bin/bash -c 'cat {self.filename} | unxz -T {self.threads} | {self.executable}'",
            shell=True,
            stdout=subprocess.PIPE,
        )

        self.counter = 0
        self.length = None

    def __len__(self):
        """
        A slightly expensive way to calculate chain lengths
        """
        if self.length is None:
            counter = subprocess.Popen(
                f"/bin/bash -c 'cat {self.filename} | unxz -T {self.threads} | {self.executable} | wc -l'",
                shell=True,
                stdout=subprocess.PIPE,
            )
            self.length = int(counter.stdout.readline().rstrip())
            counter.terminate()
            counter.wait()

        return self.length

    def terminate_child(self):
        """
        A blocking call to terminate the child
        """
        self.child.terminate()
        self.child.wait()

    def __iter__(self):
        return self

    def __next__(self):
        if self.flips:
            partition = self.read_flips()

            if partition.parent:
                partition.parent.parent = None
        else:
            partition = self.read_assignment()

        self.counter += 1
        return partition

    def read_flips(self):
        delta_line = self.child.stdout.readline()
        if not delta_line:
            self.terminate_child()
            raise StopIteration

        # assignment = ast.literal_eval(assignment_line.decode().rstrip())
        delta = json.loads(delta_line)

        if not isinstance(delta, list) or not delta:
            self.terminate_child()
            raise TypeError("Invalid chain!")

        delta_assignment = {}
        for district, nodes in enumerate(delta):  # GerryChain is 1-indexed
            for node in nodes:
                delta_assignment[node] = district+1

        if self.counter == 0:
            args = [
                self.graph,
                delta_assignment,
            ]
            if self.updaters:
                args.append(self.updaters)
            args.extend(self.args)

            if self.geographic:
                self.partition = GeographicPartition(
                    *args,
                    **self.kwargs,
                )
            else:
                self.partition = Partition(
                    *args,
                    **self.kwargs,
                )
        else:
            self.partition = self.partition.flip(delta_assignment)

        return self.partition

    def read_assignment(self):
        assignment_line = self.child.stdout.readline()
        if not assignment_line:
            self.terminate_child()
            raise StopIteration

        # assignment = ast.literal_eval(assignment_line.decode().rstrip())
        assignment = json.loads(assignment_line)

        if not isinstance(assignment, list) or not assignment:
            self.terminate_child()
            raise TypeError("Invalid chain!")

        assignment = [x+1 for x in assignment]  # GerryChain is 1-indexed

        args = [
            self.graph,
            dict(enumerate(assignment)),
        ]
        if self.updaters:
            args.append(self.updaters)
        args.extend(self.args)

        if self.geographic:
            return GeographicPartition(
                *args,
                **self.kwargs,
            )
        else:
            return Partition(
                *args,
                **self.kwargs,
            )


from typing import Optional
import pydantic

class ChainMetadata(pydantic.BaseModel):
    filename: str  # absolute path
    start_timestamp: int  # unix epoch of end time
    end_timestamp: int  # unix epoch of end time
    graph_hash: str  # sha256
    user: str

    git_commit: Optional[str]
    git_repo_clean: Optional[bool]

    ip: Optional[str]
    server_timestamp: Optional[int]
    user_agent: Optional[str]
    identifier: Optional[str]

    class Config:  # allow arbitrary user attributes
        extra = "allow"


from .utils import compute_graph_hash

from gerrychain import Partition
from typing import Iterable
import time
import json
import multiprocessing
import getpass
import os
import git
import requests

import subprocess


class Record:
    def __init__(
        self,
        chain: Iterable[Partition],
        filename,
        executable: str = "pcompress",
        # executable="pv",
        threads: int = None,
        extreme: bool = True,
        metadata: dict = {},
        cloud: bool = False,
        cloud_url: str = "http://127.0.0.1:5000",
        api_key: str = None
    ):
        self.chain = iter(chain)
        self.filename = filename
        self.extreme = extreme
        self.executable = executable
        self.cloud = cloud
        self.cloud_url = cloud_url
        self.metadata = metadata
        self.start_time = int(time.time())

        if not api_key:
            self.api_key = os.environ.get("GERRYCHAIN_API_KEY")
        else:
            self.api_key = api_key

        if not threads:
            self.threads = multiprocessing.cpu_count()
        else:
            self.threads = threads

        if self.extreme:
            if self.executable == "pcompress":
                self.executable = "pcompress -e"

            self.child = subprocess.Popen(
                f"{self.executable} | xz -e -T {self.threads} > {self.filename}",
                # f"{self.executable} | xz --lzma2=preset=9e,lp=1,lc=0,pb=0,mf=bt3 -T {self.threads} > {self.filename}",
                shell=True,
                stdin=subprocess.PIPE,
            )
        else:
            self.child = subprocess.Popen(
                f"{self.executable} | xz -T {self.threads} > {self.filename}",
                shell=True,
                stdin=subprocess.PIPE,
            )

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.chain)

    def __next__(self):
        try:
            step = next(self.chain)
            assignment = list(step.assignment.to_series().sort_index().astype(int))
            minimum = min(assignment)
            assignment = [x-minimum for x in assignment]  # GerryChain is sometimes 1-indexed

            state = str(assignment).rstrip() + "\n"
            self.child.stdin.write(state.encode())
            return step

        except StopIteration:  # kill child process
            self.end_time = int(time.time())

            self.child.stdin.close()
            self.child.wait()

            self.child.terminate()
            self.child.wait()

            if self.cloud:
                self.metadata |= {
                    "start_timestamp": self.start_time, 
                    "end_timestamp": self.end_time, 
                    "filename": self.filename, 
                    "graph_hash": compute_graph_hash(self.chain.state.graph),
                    "user": getpass.getuser(),
                    "git_commit": None,
                    "git_repo_clean": None
                    }

                try:
                    repo = git.Repo(".")
                    if repo.index.diff(repo.head.commit) or repo.untracked_files:
                        dirty = True
                        self.metadata["git_repo_clean"] = False
                    else:
                        self.metadata["git_repo_clean"] = True

                    self.metadata["commit"] = str(repo.head.commit)
                except (git.exc.InvalidGitRepositoryError, ValueError) as e:
                    pass

                upload_status = requests.post(self.cloud_url, files={
                    "pcompress": open(self.filename, "rb")
                }, data = self.metadata, headers= {
                    "GERRYCHAIN-API-KEY": self.api_key
                })

                if upload_status.status_code == 200:
                    print("Chain object uploaded as:", upload_status.text)  
                elif upload_status.status_code == 401:
                    print("Access denied. Upload failed.")
                else:
                    print("Server error. Upload failed.")

            raise
